<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/6/16
 * Time: 8:02 PM
 */

namespace qfq;

use qfq;
use qfq\Store;
use qfq\OnArray;
use qfq\UserFormException;

require_once(__DIR__ . '/../qfq/store/Store.php');
require_once(__DIR__ . '/../qfq/Constants.php');
require_once(__DIR__ . '/../qfq/exceptions/DbException.php');
require_once(__DIR__ . '/../qfq/exceptions/UserFormException.php');
require_once(__DIR__ . '/../qfq/Database.php');
require_once(__DIR__ . '/../qfq/helper/HelperFormElement.php');
require_once(__DIR__ . '/../qfq/helper/Support.php');
require_once(__DIR__ . '/../qfq/helper/OnArray.php');
require_once(__DIR__ . '/../qfq/report/Link.php');


/**
 * Class AbstractBuildForm
 * @package qfq
 */
abstract class AbstractBuildForm {
    protected $formSpec = array();  // copy of the loaded form
    protected $feSpecAction = array(); // copy of all formElement.class='action' of the loaded form
    protected $feSpecNative = array(); // copy of all formElement.class='native' of the loaded form
    protected $buildElementFunctionName = array();
    protected $pattern = array();
    protected $wrap = array();
    protected $symbol = array();
    protected $showDebugInfo = false;
    protected $inputCheckPattern = array();

//    protected $feDivClass = array(); // Wrap FormElements in <div class="$feDivClass[type]">

    /**
     * @var Store
     */
    protected $store = null;
    /**
     * @var Evaluate
     */
    protected $evaluate = null;
    /**
     * @var string
     */
    private $formId = null;
    /**
     * @var Sip
     */
    private $sip = null;

    /**
     * AbstractBuildForm constructor.
     *
     * @param array $formSpec
     * @param array $feSpecAction
     * @param array $feSpecNative
     */
    public function __construct(array $formSpec, array $feSpecAction, array $feSpecNative) {
        $this->formSpec = $formSpec;
        $this->feSpecAction = $feSpecAction;
        $this->feSpecNative = $feSpecNative;
        $this->store = Store::getInstance();
        $this->db = new Database();
        $this->evaluate = new Evaluate($this->store, $this->db);
        $this->showDebugInfo = ($this->store->getVar(SYSTEM_SHOW_DEBUG_INFO, STORE_SYSTEM) === 'yes');

        $this->sip = $this->store->getSipInstance();

        // render mode specific
        $this->fillWrap();

        $this->buildElementFunctionName = [
            'checkbox' => 'Checkbox',
            'date' => 'DateTime',
            'datetime' => 'DateTime',
            'dateJQW' => 'DateJQW',
            'datetimeJQW' => 'DateJQW',
            'email' => 'Input',
            'gridJQW' => 'GridJQW',
            FE_TYPE_EXTRA => 'Extra',
            'text' => 'Input',
            'editor' => 'Editor',
            'time' => 'DateTime',
            'note' => 'Note',
            'password' => 'Input',
            'radio' => 'Radio',
            'select' => 'Select',
            'subrecord' => 'Subrecord',
            'upload' => 'File',
            'fieldset' => 'Fieldset',
            'pill' => 'Pill',
            'templateGroup' => 'TemplateGroup'
        ];

        $this->buildRowName = [
            'checkbox' => 'Native',
            'date' => 'Native',
            'datetime' => 'Native',
            'dateJQW' => 'Native',
            'datetimeJQW' => 'Native',
            'email' => 'Native',
            'gridJQW' => 'Native',
            FE_TYPE_EXTRA => 'Native',
            'text' => 'Native',
            'editor' => 'Native',
            'time' => 'Native',
            'note' => 'Native',
            'password' => 'Native',
            'radio' => 'Native',
            'select' => 'Native',
            'subrecord' => 'Subrecord',
            'upload' => 'Native',
            'fieldset' => 'Fieldset',
            'pill' => 'Pill',
            'templateGroup' => 'TemplateGroup'
        ];

        $this->symbol[SYMBOL_EDIT] = "<span class='glyphicon " . GLYPH_ICON_EDIT . "'></span>";
        $this->symbol[SYMBOL_NEW] = "<span class='glyphicon " . GLYPH_ICON_NEW . "'></span>";
        $this->symbol[SYMBOL_DELETE] = "<span class='glyphicon " . GLYPH_ICON_DELETE . "'></span>";

        $this->inputCheckPattern = Sanitize::inputCheckPatternArray();
    }

    abstract public function fillWrap();

    /**
     * Builds complete form. Depending of form specification, the layout will be 'plain' / 'table' / 'bootstrap'.
     *
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string|array   $mode=LOAD_FORM: The whole form as HTML, $mode=FORM_UPDATE: array of all formElement.dynamicUpdate-yes  values/states
     * @throws CodeException
     * @throws DbException
     * @throws \qfq\UserFormException
     */
    public function process($mode, $htmlElementNameIdZero = false) {
        $htmlHead = '';
        $htmlTail = '';
        $htmlT3vars = '';
        $htmlSubrecords = '';
        $htmlElements = '';
        $json = array();

        $modeCollectFe = FLAG_DYNAMIC_UPDATE;
        $storeUse = STORE_USE_DEFAULT;

        if ($mode === FORM_SAVE) {
            $modeCollectFe = FLAG_ALL;
            $storeUse = STORE_RECORD . STORE_TABLE_DEFAULT;
        }

        // <form>
        if ($mode === FORM_LOAD) {
            $htmlHead = $this->head();
        }

        $filter = $this->getProcessFilter();

        if ($this->formSpec['multiMode'] !== 'none') {

            $parentRecords = $this->db->sql($this->formSpec['multiSql']);
            foreach ($parentRecords as $row) {
                $this->store->setStore($row, STORE_PARENT_RECORD, true);
                $jsonTmp = array();
                $htmlElements = $this->elements($row['_id'], $filter, 0, $jsonTmp, $modeCollectFe);
                $json[] = $jsonTmp;
            }
        } else {
            $recordId = $this->store->getVar(SIP_RECORD_ID, STORE_SIP);
            $htmlElements = $this->elements($recordId, $filter, 0, $json, $modeCollectFe, $htmlElementNameIdZero, $storeUse, $mode);
        }

        // <form>
        if ($mode === FORM_LOAD) {
            $htmlT3vars = $this->prepareT3VarsForSave();
            $htmlTail = $this->tail();
            $htmlSubrecords = $this->doSubrecords();
        }

        $htmlSip = $this->buildHiddenSip($json);

        return ($mode === FORM_LOAD) ? $htmlHead . $htmlElements . $htmlSip . $htmlT3vars . $htmlTail . $htmlSubrecords : $json;
    }

    /**
     * Builds the head area of the form.
     *
     * @return string
     */
    public function head() {
        $html = '';

        $html .= '<div ' . Support::doAttribute('class', $this->formSpec[F_CLASS], TRUE) . '>'; // main <div class=...> around everything

        // Logged in BE User will see a FormEdit Link
        $sipParamString = OnArray::toString($this->store->getStore(STORE_SIP), ':', ', ', "'");
        $formEditUrl = $this->createFormEditUrl();

        $html .= "<p><a " . Support::doAttribute('href', $formEditUrl) . ">Edit</a> <small>[$sipParamString]</small></p>";

        $html .= $this->wrapItem(WRAP_SETUP_TITLE, $this->formSpec['title'], true);

        $html .= $this->getFormTag();

        return $html;
    }

    /**
     * If SHOW_DEBUG_INFO=yes: create a link (incl. SIP) to edit the current form. Show also the hidden content of the SIP.
     *
     * @return string String: <a href="?pageId&sip=....">Edit</a> <small>[sip:..., r:..., urlparam:..., ...]</small>
     */
    public function createFormEditUrl() {

        if (!$this->showDebugInfo) {
            return '';
        }

        $queryStringArray = [
            'id' => $this->store->getVar(TYPO3_PAGE_ID, STORE_TYPO3),
            'form' => 'form',
            'r' => $this->formSpec['id']
        ];

        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();
        $url = $sip->queryStringToSip($queryString);

        return $url;
    }

    /**
     * Wrap's $this->wrap[$item][WRAP_SETUP_START] around $value. If $flagOmitEmpty==true && $value=='': return ''.
     *
     * @param $item
     * @param $value
     * @param bool|false $flagOmitEmpty
     * @return string
     */
    public function wrapItem($item, $value, $flagOmitEmpty = false) {

        if ($flagOmitEmpty && $value === "") {
            return '';
        }

        return $this->wrap[$item][WRAP_SETUP_START] . $value . $this->wrap[$item][WRAP_SETUP_END];
    }

    /**
     * Returns '<form ...>'-tag with various attributes.
     *
     * @return string
     */
    public function getFormTag() {

        $attribute = $this->getFormTagAtrributes();

        return '<form ' . OnArray::toString($attribute, '=', ' ', "'") . '>';
    }

    /**
     * Build an assoc array with standard form attributes.
     *
     * @return array
     */
    public function getFormTagAtrributes() {

        $attribute['id'] = $this->getFormId();
        $attribute['method'] = 'post';
        $attribute['action'] = $this->getActionUrl();
        $attribute['target'] = '_top';
        $attribute['accept-charset'] = 'UTF-8';
        $attribute['autocomplete'] = 'on';
        $attribute['enctype'] = $this->getEncType();

        return $attribute;
    }

    /**
     * Return a uniq form id
     *
     * @return string
     */
    public function getFormId() {
        if ($this->formId === null) {
            $this->formId = uniqid('qfq-form-');
        }
        return $this->formId;
    }

    /**
     * Builds the HTML 'form'-tag inlcuding all attributes and target.
     *
     * Notice: the SIP will be transferred as POST Parameter.
     *
     * @return string
     * @throws DbException
     */
    public function getActionUrl() {

        return API_DIR . '/save.php';
    }

    /**
     * Determines the enctype.
     *
     * See: https://www.w3.org/wiki/HTML/Elements/form#HTML_Attributes
     *
     * @return string
     * @throws DbException
     */
    public function getEncType() {

        $result = $this->db->sql("SELECT id FROM FormElement AS fe WHERE fe.formId=? AND fe.type='upload' LIMIT 1", ROW_REGULAR, [$this->formSpec['id']], 'Look for Formelement.type="upload"');
        return (count($result) === 1) ? 'multipart/form-data' : 'application/x-www-form-urlencoded';

    }

    abstract public function getProcessFilter();

    /**
     * Process all FormElements: build corresponding HTML code. Collect and return all HTML code.
     *
     * @param $recordId
     * @param string $filter FORM_ELEMENTS_NATIVE | FORM_ELEMENTS_SUBRECORD | FORM_ELEMENTS_NATIVE_SUBRECORD
     * @param int $feIdContainer
     * @param array $json
     * @param string $modeCollectFe
     * @param bool $htmlElementNameIdZero
     * @param string $storeUse
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws CodeException
     * @throws DbException
     * @throws \qfq\UserFormException
     */
    public function elements($recordId, $filter = FORM_ELEMENTS_NATIVE, $feIdContainer = 0, array &$json,
                             $modeCollectFe = FLAG_DYNAMIC_UPDATE, $htmlElementNameIdZero = false, $storeUse = STORE_USE_DEFAULT, $mode = FORM_LOAD) {
        $html = '';
        $flagOutput = false;

        // The following 'FormElement.parameter' will never be used during load (fe.type='upload').
        $skip = [FE_SQL_UPDATE, FE_SQL_INSERT, FE_SQL_DELETE, FE_SQL_AFTER, FE_SQL_BEFORE, F_FE_PARAMETER];

        // get current data record
        if ($recordId > 0 && $this->store->getVar('id', STORE_RECORD) === false) {
            $row = $this->db->sql("SELECT * FROM " . $this->formSpec[F_TABLE_NAME] . " WHERE id = ?", ROW_EXPECT_1, array($recordId), "Form '" . $this->formSpec[F_NAME] . "' failed to load record '$recordId' from table '" . $this->formSpec[F_TABLE_NAME] . "'.");
            $this->store->setStore($row, STORE_RECORD);
        }

        $this->checkAutoFocus();

        // Iterate over all FormElements
        foreach ($this->feSpecNative as $fe) {

            if (($filter === FORM_ELEMENTS_NATIVE && $fe[FE_TYPE] === 'subrecord')
                || ($filter === FORM_ELEMENTS_SUBRECORD && $fe[FE_TYPE] !== 'subrecord')
//                || ($filter === FORM_ELEMENTS_DYNAMIC_UPDATE && $fe[FE_DYNAMIC_UPDATE] === 'no')
            ) {
                continue; // skip this FE
            }

            $flagOutput = ($fe[FE_TYPE] !== FE_TYPE_EXTRA); // type='extra' will not displayed not trasnmitted to the form

            $debugStack = array();

            // Preparation for Log, Debug
            $this->store->setVar(SYSTEM_FORM_ELEMENT, Logger::formatFormElementName($fe), STORE_SYSTEM);

            // for Upload FormElements, it's necessary to precalculate an optional given 'slaveId'.
            if ($fe[FE_TYPE] === FE_TYPE_UPLOAD) {
                Support::setIfNotSet($fe, FE_SLAVE_ID);
                $slaveId = Support::falseEmptyToZero($this->evaluate->parse($fe[FE_SLAVE_ID]));
                $this->store->setVar(VAR_SLAVE_ID, $slaveId, STORE_VAR);
            }

            // evaluate current FormElement
            $formElement = $this->evaluate->parseArray($fe, $skip, $debugStack);

            // Some Defaults
            $formElement = Support::setFeDefaults($formElement, $this->formSpec);

            if ($flagOutput === true) {
                $this->fillWrapLabelInputNote($formElement[FE_BS_LABEL_COLUMNS], $formElement[FE_BS_INPUT_COLUMNS], $formElement[FE_BS_NOTE_COLUMNS]);
            }

            //In case the current element is a 'RETYPE' element: take the element name of the source FormElement. Needed in the next row to retrieve the default value.
            $name = (isset($formElement[FE_RETYPE_SOURCE_NAME])) ? $formElement[FE_RETYPE_SOURCE_NAME] : $formElement[FE_NAME];

            // Get default value
//            $value = ($formElement[FE_VALUE] === '') ? $this->store->getVar($name, $storeUse,
//                $formElement['checkType']) : $formElement[FE_VALUE];

            // If there is a value explicit defined: take it
            $value = $formElement[FE_VALUE];
            //
            if ($value === '') {
                // Only take the default, if the FE is a real tablecolumn.
                if ($this->store->getVar($formElement[FE_NAME], STORE_TABLE_COLUMN_TYPES) !== false) {
                    $value = $this->store->getVar($name, $storeUse, $formElement['checkType']);
                }
            }

            // Typically: $htmlElementNameIdZero = true
            // After Saving a record, staying on the form, the FormElements on the Client are still known as '<feName>:0'.
            $htmlFormElementId = HelperFormElement::buildFormElementName($formElement[FE_NAME], ($htmlElementNameIdZero) ? 0 : $recordId);

            // Construct Marshaller Name: buildElement
            $buildElementFunctionName = 'build' . $this->buildElementFunctionName[$formElement[FE_TYPE]];

            $jsonElement = array();
            // Render pure element
            $elementHtml = $this->$buildElementFunctionName($formElement, $htmlFormElementId, $value, $jsonElement, $mode);

            // container elements do not have dynamicUpdate='yes'. Instead they deliver nested elements.
            if ($formElement['class'] == 'container') {
                if (count($jsonElement) > 0) {
                    $json = array_merge($json, $jsonElement);
                }
            } else {
                // for non container elements: just add the current json status
                if ($modeCollectFe === FLAG_ALL || ($modeCollectFe == FLAG_DYNAMIC_UPDATE && $fe[FE_DYNAMIC_UPDATE] == 'yes')) {
                    if (isset($jsonElement[0]) && is_array($jsonElement[0])) {
                        // Checkboxes are delivered as array of arrays: unnest them and append them to the existing json array.
                        $json = array_merge($json, $jsonElement);
                    } else {
                        $json[] = $jsonElement;
                    }
                }
            }

            if ($flagOutput) {
                // debugStack as Tooltip
                if ($this->showDebugInfo && count($debugStack) > 0) {
                    $elementHtml = Support::appendTooltip($elementHtml, implode("\n", $debugStack));
                }

                // Construct Marshaller Name: buildRow
                $buildRowName = 'buildRow' . $this->buildRowName[$formElement[FE_TYPE]];

                $html .= $formElement[FE_HTML_BEFORE] . $this->$buildRowName($formElement, $elementHtml, $htmlFormElementId) . $formElement[FE_HTML_AFTER];
            }
        }

        // Log / Debug: Last FormElement has been processed.
        $this->store->setVar(SYSTEM_FORM_ELEMENT, '', STORE_SYSTEM);

        return $html;
    }

    /**
     * Check if there is an explicit 'autofocus' definition in at least one FE.
     * Found: do nothing, it will be rendered at the correct position.
     * Not found: set 'autofocus' on the first FE.
     *
     * Accepted misbehaviour on forms with pills: if there is at least one editable element on the first pill,
     *   the other pills are not checked - independent if there was a definition on the first pill or not.
     *   Reason: checks happens per pill - if there is no explizit definition on the first pill, take the first editable
     *           element of that pill.
     */
    private function checkAutoFocus() {
        static $found = false;
        $idx = false;

        if ($found) {
            return;
        }

        // Search if there is an explicit autofocus definition.
        for ($i = 0; $i < count($this->feSpecNative); ++$i) {
            // Only check native elements which will be shown
            if ($this->feSpecNative[$i][FE_CLASS] == FE_CLASS_NATIVE &&
                ($this->feSpecNative[$i][FE_MODE] == FE_MODE_SHOW || $this->feSpecNative[$i][FE_MODE] == FE_MODE_REQUIRED)
            ) {
                // Check if there is an explicit definition.
                if (isset($this->feSpecNative[$i][FE_AUTOFOCUS])) {
                    if ($this->feSpecNative[$i][FE_AUTOFOCUS] == '' || $this->feSpecNative[$i][FE_AUTOFOCUS] == '1') {
                        $this->feSpecNative[$i][FE_AUTOFOCUS] = '1'; // fix to '=1'
                    } else {
                        unset($this->feSpecNative[$i][FE_AUTOFOCUS]);
                    }
                    $found = true;
                    return;
                }

                if ($idx === false) {
                    $idx = $i;
                }
            }
        }

        // No explicit definition found: take the first found editable element.
        if ($idx !== false) {
            $found = true;
            // No explicit definition found: set autofocus.
            $this->feSpecNative[$idx][FE_AUTOFOCUS] = '1';
        }
    }

    abstract public function fillWrapLabelInputNote($label, $input, $note);

    /**
     * Copy a subset of current STORE_TYPO3 variables to SIP. Set a hidden form field to submit the assigned SIP to save/update.
     *
     * @throws CodeException
     * @throws UserFormException
     */
    private function prepareT3VarsForSave() {

        $t3VarsSip = $this->store->copyT3VarsToSip();

        return $this->buildNativeHidden(CLIENT_TYPO3VARS, $t3VarsSip);

    }

    /**
     * Builds a real HTML hidden form element. Useful for checkboxes, Multiple-Select and Radios.
     *
     * @param $htmlFormElementId
     * @param $value
     * @return string
     */
    public function buildNativeHidden($htmlFormElementId, $value) {
        return '<input type="hidden" name="' . $htmlFormElementId . '" value="' . htmlentities($value) . '">';
    }

    abstract public function tail();

    abstract public function doSubrecords();

    /**
     * Create a hidden sip, based on latest STORE_SIP Values. Return complete HTML 'hidden' element.
     *
     * @param array $json
     * @return string  <input type='hidden' name='s' value='<sip>'>
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildHiddenSip(array &$json) {

        $sipArray = $this->store->getStore(STORE_SIP);

        // do not include system vars
        unset($sipArray[SIP_SIP]);
        unset($sipArray[SIP_URLPARAM]);

        $queryString = Support::arrayToQueryString($sipArray);
        $sip = $this->store->getSipInstance();

        $sipValue = $sip->queryStringToSip($queryString, RETURN_SIP);

        $json[] = $this->getJsonElementUpdate(CLIENT_SIP, $sipValue, FE_MODE_SHOW);

        return $this->buildNativeHidden(CLIENT_SIP, $sipValue);
    }

    /**
     * Create an array with standard elements and add 'form-element', 'value'.
     *
     * @param $htmlFormElementId
     * @param string|array $value
     * @param string $feMode disabled|readonly|''
     * @return array
     */
    private function getJsonElementUpdate($htmlFormElementId, $value, $feMode) {

        $json = $this->getJsonFeMode($feMode);

        $json['form-element'] = $htmlFormElementId;
        $json['value'] = $value;

        return $json;
    }

    /**
     * Set corresponding JSON attributes readonly/required/disabled, based on $formElement[FE_MODE].
     *
     * @param array $feMode
     * @return array
     * @throws UserFormException
     */
    private function getJsonFeMode($feMode) {

        $this->getFeMode($feMode, $hidden, $disabled, $required);

        return [API_JSON_HIDDEN => $hidden === 'yes', API_JSON_DISABLED => $disabled === 'yes', API_JSON_REQUIRED => $required === 'yes'];
    }

    /**
     * Depending of $feMode set variables $hidden, $disabled, $required to 'yes' or 'no'.
     *
     * @param $feMode
     * @param $hidden
     * @param $disabled
     * @param $required
     * @throws \qfq\UserFormException
     */
    private function getFeMode($feMode, &$hidden, &$disabled, &$required) {
        $hidden = 'no';
        $disabled = 'no';
        $required = 'no';

        switch ($feMode) {
            case FE_MODE_SHOW:
                break;
            case FE_MODE_REQUIRED:
                $required = 'yes';
                break;
            case FE_MODE_READONLY:
                $disabled = 'yes';  // convert the UI status 'readonly' to the HTML/CSS status 'disabled'.
                break;
            case FE_MODE_HIDDEN:
                $hidden = 'yes';
                break;
            default:
                throw new UserFormException("Unknown mode '$feMode'", ERROR_UNKNOWN_MODE);
                break;
        }
    }

    /**
     * Takes the current SIP ('form' and additional parameter), set SIP_RECORD_ID=0 and create a new 'NewRecordUrl'.
     *
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function deriveNewRecordUrlFromExistingSip(&$toolTipNew) {
        $urlParam = $this->store->getStore(STORE_SIP);
        $urlParam[SIP_RECORD_ID] = 0;

        unset($urlParam[SIP_SIP]);
        unset($urlParam[SIP_URLPARAM]);

        Support::appendTypo3ParameterToArray($urlParam);

        $sip = $this->store->getSipInstance();

        $url = $sip->queryStringToSip(OnArray::toString($urlParam));

        if ($this->showDebugInfo) {
            //TODO: missing decoding of SIP
            $toolTipNew .= PHP_EOL . PHP_EOL . OnArray::toString($urlParam, ' = ', PHP_EOL, "'");
        }

        return $url;
    }

    abstract public function buildRowPill(array $formElement, $elementHtml);

    abstract public function buildRowFieldset(array $formElement, $elementHtml);

    abstract public function buildRowTemplateGroup(array $formElement, $elementHtml);

    abstract public function buildRowSubrecord(array $formElement, $elementHtml);

    /**
     * Builds a label, typically for an html-'<input>'-element.
     *
     * @param string $htmlFormElementId
     * @param string $label
     * @return string
     */
    public function buildLabel($htmlFormElementId, $label) {
        $attributes = Support::doAttribute('for', $htmlFormElementId);
        $attributes .= Support::doAttribute('class', 'control-label');

        $html = Support::wrapTag("<label $attributes>", $label);

        return $html;
    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="email|input|password|url" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws \qfq\UserFormException
     */
    public function buildInput(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $textarea = '';

        $attribute = Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');

        if (isset($formElement[FE_RETYPE_SOURCE_NAME])) {
            $htmlFormElementIdPrimary = str_replace(RETYPE_FE_NAME_EXTENSION, '', $htmlFormElementId);
            $attribute .= Support::doAttribute('data-match', '[name=' . str_replace(':', '\\:', $htmlFormElementIdPrimary) . ']');
        }

        // Check for input type 'textarea'
        $colsRows = explode(',', $formElement['size'], 2);
        if (count($colsRows) === 2) {
            // <textarea>
            $htmlTag = '<textarea';

            $attribute .= Support::doAttribute('cols', $colsRows[0]);
            $attribute .= Support::doAttribute('rows', $colsRows[1]);
            $textarea = htmlentities($value) . '</textarea>';

        } else {
            $htmlTag = '<input';

            $this->adjustMaxLength($formElement);

            if ($formElement['maxLength'] > 0 && $value !== '') {
                // crop string only if it's not empty (substr returns false on empty strings)
                $value = substr($value, 0, $formElement['maxLength']);
            }

            // 'maxLength' needs an upper 'L': naming convention for DB tables!
            $attribute .= $this->getAttributeList($formElement, ['type', 'size', 'maxLength']);
            $attribute .= Support::doAttribute('value', htmlentities($value), false);
        }

        $attribute .= $this->getAttributeList($formElement, ['autocomplete', 'autofocus', 'placeholder']);
        $attribute .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getInputCheckPattern($formElement['checkType'], $formElement['checkPattern']);

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return "$htmlTag $attribute>$textarea" . $this->getHelpBlock();

    }

    /**
     * Calculates the maxlength of an input field, based on formElement type, formElement user definition and table.field definition.
     *
     * @param array $formElement
     */
    private function adjustMaxLength(array &$formElement) {

        // MIN( $formElement['maxLength'], tabledefinition)
        $maxLength = $this->getColumnSize($formElement['name']);

        switch ($formElement[FE_TYPE]) {
            case 'date':
                $feMaxLength = 10;
                break;
            case 'datetime':
                $feMaxLength = 19;
                break;
            case 'time':
                $feMaxLength = 8;
                break;
            default:
                $feMaxLength = false;
                break;
        }

        // In case the underlying tablecolumn is not of type date/time, the $maxLength might be to high: correct
        if ($feMaxLength !== false && $maxLength !== false && $feMaxLength < $maxLength) {
            $maxLength = $feMaxLength;
        }

        // date/datetime
        if ($maxLength !== false) {
            if (is_numeric($formElement['maxLength']) && $formElement['maxLength'] != 0) {
                if ($formElement['maxLength'] > $maxLength) {
                    $formElement['maxLength'] = $maxLength;
                }
            } else {
                $formElement['maxLength'] = $maxLength;
            }
        }
    }

    /**
     * Get column spec from tabledefinition and parse size of it. If nothing defined, return false.
     *
     * @param $column
     * @return bool|int  a) 'false' if there is no length definition, b) length definition, c) date|time|datetime|timestamp use hardcoded length
     */
    private function getColumnSize($column) {
        $matches = array();

        $typeSpec = $this->store->getVar($column, STORE_TABLE_COLUMN_TYPES);
        switch ($typeSpec) {
            case 'date': // yyyy-mm-dd
                return 10;
            case 'datetime': // yyyy-mm-dd hh:mm:ss
            case 'timestamp': // yyyy-mm-dd hh:mm:ss
                return 19;
            case 'time': // hh:mm:ss
                return 8;
            default:
                if (substr($typeSpec, 0, 4) === 'set(' || substr($typeSpec, 0, 5) === 'enum(') {
                    return $this->maxLengthSetEnum($typeSpec);
                }
                break;
        }

        // e.g.: string(64) >> 64, enum('yes','no') >> false
        if (1 === preg_match('/\((.+)\)/', $typeSpec, $matches)) {
            if (is_numeric($matches[1]))
                return $matches[1];
        }

        return false;
    }

    /**
     * Get the strlen of the longest element in enum('val1','val2',...,'valn') or set('val1','val2',...,'valn')
     *
     * @param string $typeSpec
     * @return int
     */
    private function maxLengthSetEnum($typeSpec) {
        $startPos = (substr($typeSpec, 0, 4) === 'set(') ? 4 : 5;
        $max = 0;

        $valueList = substr($typeSpec, $startPos, strlen($typeSpec) - $startPos - 1);
        $valueArr = explode(',', $valueList);
        foreach ($valueArr as $value) {
            $value = trim($value, "'");
            $len = strlen($value);
            if ($len > $max) {
                $max = $len;
            }
        }

        return $max;
    }

    /**
     * Builds a HTML attribute list, based on  $attributeList.
     *
     * E.g.: attributeList: [ 'type', 'autofocus' ]
     *       generates: 'type="$formElement[FE_TYPE]" autofocus="$formElement['autofocus']" '
     *
     * @param array $formElement
     * @param array $attributeList
     * @param bool $flagOmitEmpty
     * @return string
     */
    private function getAttributeList(array $formElement, array $attributeList, $flagOmitEmpty = true) {
        $attribute = '';
        foreach ($attributeList as $item) {
            if (isset($formElement[$item]))
                $attribute .= Support::doAttribute(strtolower($item), $formElement[$item], $flagOmitEmpty);
        }
        return $attribute;
    }

    /**
     * Construct HTML Input attribute for Client Validation:
     *
     *   type     data                      result
     *   -------  -----------------------   -------------------------------------------------------------------------------
     *   min|max  <min value>|<max value>   min="$attrData[0]"|max="$attrData[1]"
     *   pattern  <regexp>                  pattern="$data"
     *   digit    -                         pattern="^[0-9]*$"
     *   email    -                         pattern="^[_a-z0-9-]+(\.[_a-z0-9-]+)*@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,4})$"
     *   alnumx   -
     *
     * For 'min/max' and 'pattern' the 'data' will be injected in the attribute string via '%s'.
     *
     * @param $type
     * @param $data
     * @return string
     * @throws \qfq\UserFormException
     */
    private function getInputCheckPattern($type, $data) {
        $attribute = '';

        if ($type === '') {
            return '';
        }

        switch ($type) {
            case SANITIZE_ALLOW_MIN_MAX:
            case SANITIZE_ALLOW_MIN_MAX_DATE:
                $arrData = explode("|", $data);
                if (count($arrData) != 2 || $arrData[0] == '' || $arrData[1] == '')
                    throw new UserFormException("Missing MIN|MAX values", ERROR_MISSING_MIN_MAX);

                $attribute = 'min="' . $arrData[0] . '" ';
                $attribute .= 'max="' . $arrData[1] . '" ';
                break;

            case SANITIZE_ALLOW_PATTERN:
                $attribute = 'pattern="' . $data . '" ';
                break;

            case SANITIZE_ALLOW_ALL:
                break;

            default:
                $attribute = 'pattern="' . $this->inputCheckPattern[$type] . '" ';
                break;
        }

        return $attribute;
    }

    /**
     * Set corresponding html attributes readonly/required/disabled, based on $formElement[FE_MODE].
     *
     * @param string $feMode
     * @return string
     * @throws UserFormException
     */
    private function getAttributeFeMode($feMode) {
        $attribute = '';

        $this->getFeMode($feMode, $hidden, $disabled, $required);

        switch ($feMode) {
            case FE_MODE_HIDDEN:
            case FE_MODE_SHOW:
                break;
            case FE_MODE_REQUIRED:
            case FE_MODE_READONLY:
                $attribute .= Support::doAttribute($feMode, $feMode);
                break;
            default:
                throw new UserFormException("Unknown mode '$feMode'", ERROR_UNKNOWN_MODE);
                break;
        }

        // Attributes: data-...
        $attribute .= Support::doAttribute(DATA_HIDDEN, $hidden);
        $attribute .= Support::doAttribute(DATA_DISABLED, $disabled);
        $attribute .= Support::doAttribute(DATA_REQUIRED, $required);

        return $attribute;
    }

    /**
     * Build HelpBlock
     *
     * @return string
     */
    private function getHelpBlock() {
        return '<div class="help-block with-errors hidden"></div>';
    }

    /**
     * Builds HTML 'checkbox' element.
     *
     * Checkboxes will only be submitted, if they are checked. Therefore, a hidden element with the unchecked value will be transferred first.
     *
     * Format: <input type="hidden" name="$htmlFormElementId" value="$valueUnChecked">
     *         <input name="$htmlFormElementId" type="checkbox" [autofocus="autofocus"]
     *            [required="required"] [disabled="disabled"] value="<value>" [checked="checked"] >
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE*
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildCheckbox(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        // Get fallback, if 'checkBoxMode' is not defined:
        if (!isset($formElement['checkBoxMode'])) {
            // This fallback is problematic if 'set' or 'enum' has 2 : defaults to single but maybe multi is meant.
            $formElement['checkBoxMode'] = (count($itemKey) > 2) ? 'multi' : 'single';
        }

        if ($formElement['checkBoxMode'] === 'multi') {
//            $htmlFormElementId .= '[]';
        } else {
            // Fill meaningfull defaults to parameter: checked|unchecked  (CHECKBOX_VALUE_CHECKED|CHECKBOX_VALUE_UNCHECKED)
            $this->prepareCheckboxCheckedUncheckedValue($itemKey, $formElement);
        }

        $attributeBase = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attributeBase .= Support::doAttribute('type', $formElement[FE_TYPE]);
        $attributeBase .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        switch ($formElement['checkBoxMode']) {
            case 'single':
                $html = $this->buildCheckboxSingle($formElement, $htmlFormElementId, $attributeBase, $value, $json);
                break;
            case 'multi';
                $html = $this->buildCheckboxMulti($formElement, $htmlFormElementId, $attributeBase, $value, $itemKey, $itemValue, $json);
                break;
            default:
                throw new UserFormException('checkBoxMode: \'' . $formElement['checkBoxMode'] . '\' is unknown.', ERROR_CHECKBOXMODE_UNKNOWN);
        }

        return $html;
    }

    /**
     * Look for key/value list (in this order, first match counts) in
     *  a) `sql1`
     *  b) `parameter:itemList`
     *  c) table.column definition
     *
     * Copies the found keys to &$itemKey and the values to &$itemValue
     * If there are no &$itemKey, copy &$itemValue to &$itemKey.
     *
     * @param array $formElement
     * @param $itemKey
     * @param $itemValue
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function getKeyValueListFromSqlEnumSpec(array $formElement, &$itemKey, &$itemValue) {
        $fieldType = '';
        $itemKey = array();
        $itemValue = array();

        if (count($formElement) < 20)
            throw new CodeException("Invalid (none or to small) Formelement", ERROR_MISSING_FORMELEMENT);

        // Call getItemsForEnumOrSet() only if there a corresponding column really exist.
        if (false !== $this->store->getVar($formElement['name'], STORE_TABLE_COLUMN_TYPES)) {
            $itemValue = $this->getItemsForEnumOrSet($formElement['name'], $fieldType);
        }

        if (is_array($formElement['sql1'])) {
            if (count($formElement['sql1']) > 0) {
                $keys = array_keys($formElement['sql1'][0]);
                $itemKey = array_column($formElement['sql1'], 'id');

                // If there is no column 'id' and at least two columns in total
                if (count($itemKey) === 0 && count($keys) >= 2) {
                    $itemKey = array_column($formElement['sql1'], $keys[0]);
                }

                $itemValue = array_column($formElement['sql1'], 'label');
                // If there is no column 'label' (e.g.: SHOW tables)
                if (count($itemValue) === 0) {
                    $idx = count($keys) == 1 ? 0 : 1;
                    $itemValue = array_column($formElement['sql1'], $keys[$idx]);
                }
            }
        } elseif (isset($formElement['itemList']) && strlen($formElement['itemList']) > 0) {
            $arr = KeyValueStringParser::parse($formElement['itemList'], ':', ',', KVP_IF_VALUE_EMPTY_COPY_KEY);
            $itemValue = array_values($arr);
            $itemKey = array_keys($arr);
        } elseif ($fieldType === 'enum' || $fieldType === 'set') {
            // already done at the beginning with '$this->getItemsForEnumOrSet($formElement['name'], $fieldType);'
        } else {
            throw new UserFormException("Missing definition (- nothing found in 'sql1', 'parameter:itemValues', 'enum-' or 'set-definition'", ERROR_MISSING_ITEM_VALUES);
        }

        if (count($itemKey) === 0) {
            $itemKey = $itemValue;
        }

        if (isset($formElement['emptyItemAtStart'])) {
            array_unshift($itemKey, '');
            array_unshift($itemValue, '');
        }

        if (isset($formElement['emptyItemAtEnd'])) {
            $itemValue[] = '';
            $itemKey[] = '';
        }

        if (isset($formElement['emptyHide'])) {
            if (isset($itemValue['']))
                unset($itemValue['']);
            if (isset($itemKey['']))
                unset($itemKey['']);

        }
    }

    /**
     * Get the attribute definition list of an enum or set column. For strings, get the default value. Return elements as an array.
     *
     * @param $column
     * @param $fieldType
     * @return array
     * @throws UserFormException
     */
    private function getItemsForEnumOrSet($column, &$fieldType) {

        // Get column definition
        $fieldTypeDefinition = $this->store->getVar($column, STORE_TABLE_COLUMN_TYPES);

        if ($fieldTypeDefinition === false) {
            throw new UserFormException("Column '$column' unknown in table '" . $this->formSpec[F_TABLE_NAME] . "'", ERROR_DB_UNKNOWN_COLUMN);
        }

        $length = strlen($fieldTypeDefinition);

        // enum('...   set('
        switch (substr($fieldTypeDefinition, 0, 4)) {
            case 'enum':
                $startPosition = 5;
                break;
            case 'set(':
                $startPosition = 4;
                break;
            default:
                $fieldType = 'string';
                return array();
        }

        // enum('a','b','c', ...)   >> [ 'a', 'b', 'c', ... ]
        // set('a','b','c', ...)   >> [ 'a', 'b', 'c', ... ]
        $values = substr($fieldTypeDefinition, $startPosition, $length - $startPosition - 1);
        $items = OnArray::trimArray(explode(',', $values), "'");
        $fieldType = substr($fieldTypeDefinition, 0, $startPosition - 1);

        return $items;
    }

    /**
     * For CheckBox's with only one checkbox: if no parameter:checked|unchecked is defined, take defaults:
     *
     *    checked: first Element in $itemKey
     *  unchecked: ''
     *
     * @param array $itemKey
     * @param array $formElement
     * @throws UserFormException
     */
    private function prepareCheckboxCheckedUncheckedValue(array $itemKey, array &$formElement) {

        if (!isset($formElement[CHECKBOX_VALUE_CHECKED])) {
            if (isset($itemKey[0])) {
                // First element in $itemKey list
                $formElement[CHECKBOX_VALUE_CHECKED] = $itemKey[0];
            } else {
                // Take column default value
                $formElement[CHECKBOX_VALUE_CHECKED] = $this->store->getVar($formElement['name'], STORE_TABLE_DEFAULT);
            }
        }

        // unchecked
        if (!isset($formElement[CHECKBOX_VALUE_UNCHECKED])) {
            if (isset($itemKey[1])) {
                $formElement[CHECKBOX_VALUE_UNCHECKED] = ($itemKey[0] === $formElement['checked']) ? $itemKey[1] : $itemKey[0];
            } else {
                $formElement[CHECKBOX_VALUE_UNCHECKED] = '';
            }
        }

        if ($formElement[CHECKBOX_VALUE_CHECKED] === $formElement[CHECKBOX_VALUE_UNCHECKED]) {
            throw new UserFormException('FormElement: type=checkbox - checked and unchecked can\'t be the same: ' . $formElement[CHECKBOX_VALUE_CHECKED], ERROR_CHECKBOX_EQUAL);
        }

    }

    /**
     * Build a Checkbox based on two values.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $attribute
     * @param $value
     * @param array $json
     * @return string
     */
    public function buildCheckboxSingle(array $formElement, $htmlFormElementId, $attribute, $value, array &$json) {
        $html = '';
        $valueJson = false;

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('value', $formElement['checked'], false);
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');

        if ($formElement['checked'] === $value) {
            $attribute .= Support::doAttribute('checked', 'checked');
            $valueJson = true;
        }

        $attribute .= $this->getAttributeList($formElement, ['autofocus']);
        $attribute .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        $html = $this->buildNativeHidden($htmlFormElementId, $formElement['unchecked']);

        $html .= '<input ' . $attribute . '>';
        if (isset($formElement['label2'])) {
            $html .= $formElement['label2'];
        }

        $html = Support::wrapTag("<label>", $html, true);
        $html = Support::wrapTag("<div class='checkbox'>", $html, true);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $valueJson, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * Build as many Checkboxes as items.
     *
     * Layout: The Bootstrap Layout needs very special setup, the checkboxes are wrapped differently with <div class=checkbox>
     *         depending of if they aligned horizontal or vertical.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $attributeBase
     * @param $value
     * @param array $itemKey
     * @param array $itemValue
     * @param array $json
     * @return string
     */
    public function buildCheckboxMulti(array $formElement, $htmlFormElementId, $attributeBase, $value, array $itemKey, array $itemValue, array &$json) {
        $json = array();

        // Defines which of the checkboxes will be checked.
        $values = explode(',', $value);

//        $attributeBase .= Support::doAttribute('name', $htmlFormElementId);
        $attributeBase .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attributeBase .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        $html = $this->buildNativeHidden(HelperFormElement::prependFormElementIdCheckBoxMulti($htmlFormElementId, 'h'), '');

        $orientation = ($formElement['maxLength'] > 1) ? ALIGN_HORIZONTAL : ALIGN_VERTICAL;
        $checkboxClass = ($orientation === ALIGN_HORIZONTAL) ? 'checkbox-inline' : 'checkbox';
        $br = '';

        $flagFirst = true;
        for ($ii = 0, $jj = 1; $ii < count($itemKey); $ii++, $jj++) {
            $jsonValue = false;
            $attribute = $attributeBase;
            $htmlFormElementIdUniq = HelperFormElement::prependFormElementIdCheckBoxMulti($htmlFormElementId, $ii);
            $attribute .= Support::doAttribute('name', $htmlFormElementIdUniq);

            // Do this only the first round.
            if ($flagFirst) {
                $flagFirst = false;
                if (isset($formElement['autofocus']))
                    $attribute .= Support::doAttribute('autofocus', $formElement['autofocus']);
            }

            $attribute .= Support::doAttribute('value', $itemKey[$ii]);

            // Check if the given key is found in field.
            if (false !== array_search($itemKey[$ii], $values)) {
                $attribute .= Support::doAttribute('checked', 'checked');
                $jsonValue = true;
            }

            // '&nbsp;' - This is necessary to correctly align an empty input.
            $value = ($itemValue[$ii] === '') ? '&nbsp;' : $itemValue[$ii];

            $htmlElement = '<input ' . $attribute . '>' . $value;

            // With ALIGN_HORIZONTAL: the label causes some trouble: skip it
            if (($orientation === ALIGN_VERTICAL)) {
                $htmlElement = Support::wrapTag('<label>', $htmlElement);
            }

            $htmlElement = Support::wrapTag("<div class='$checkboxClass'>", $htmlElement, true);

            // control orientation
            if ($formElement['maxLength'] > 1) {

                if ($jj == $formElement['maxLength']) {
                    $jj = 0;
                    $br = '<br>';
                } else {
                    $br = '';
                }
            }

            $html .= $htmlElement . $br;
            $json[] = $this->getJsonElementUpdate($htmlFormElementIdUniq, $jsonValue, $formElement[FE_MODE]);

        }

//        if (isset($formElement[CHECKBOX_ORIENTATION]) && $formElement[CHECKBOX_ORIENTATION] !== 'vertical')
//            $html = Support::wrapTag("<div class='checkbox'>", $html, true);
//
        return $html;
    }

    /**
     * Submit extra (hidden) values by SIP.
     *
     * Sometimes, it's usefull to precalculate values during formload and to submit them as hidden fields.
     * To avoid any manipulation on those fields, the values will be transferred by SIP.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildExtra(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {

        if ($mode === FORM_LOAD) {
            // Split 'grId:0' in 'grId' and '0'
            $name = explode(':', $htmlFormElementId, 2);

            $this->store->setVar($name[0], $value, STORE_SIP, false);
        }
    }

    /**
     * Build HTML 'radio' element.
     *
     * Checkboxes will only be submitted, if they are checked. Therefore, a hidden element with the unchecked value will be transfered first.
     *
     * Format: <input type="hidden" name="$htmlFormElementId" value="$valueUnChecked">
     *         <input name="$htmlFormElementId" type="radio" [autofocus="autofocus"]
     *            [required="required"] [disabled="disabled"] value="<value>" [checked="checked"] >
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildRadio(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        $attributeBase = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attributeBase .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);
        $attributeBase .= Support::doAttribute('name', $htmlFormElementId);
        $attributeBase .= Support::doAttribute('type', $formElement[FE_TYPE]);
        $attributeBase .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');

        $jj = 0;
        $flagFirst = true;

        $orientation = ($formElement['maxLength'] > 1) ? ALIGN_HORIZONTAL : ALIGN_VERTICAL;
        $radioClass = ($orientation === ALIGN_HORIZONTAL) ? 'radio-inline' : 'radio';
        $br = '';

        $html = $this->buildNativeHidden($htmlFormElementId, $value);

        for ($ii = 0; $ii < count($itemValue); $ii++) {
            $jj++;
            $attribute = $attributeBase; //
            if ($flagFirst) {
                $flagFirst = false;
                if (isset($formElement['autofocus']))
                    $attribute .= Support::doAttribute('autofocus', $formElement['autofocus']);
            }

            $attribute .= Support::doAttribute('value', $itemKey[$ii]);
            $attribute .= Support::doAttribute('title', $formElement['tooltip']);

            if ($itemKey[$ii] === $value) {
                $attribute .= Support::doAttribute('checked', 'checked');
            }

            // '&nbsp;' - This is necessary to correctly align an empty input.
            $tmpValue = ($itemValue[$ii] === '') ? '&nbsp;' : $itemValue[$ii];

            $htmlElement = '<input ' . $attribute . '>' . $tmpValue;

            // With ALIGN_HORIZONTAL: the label causes some trouble: skip it
            if (($orientation === ALIGN_VERTICAL)) {
                $htmlElement = Support::wrapTag('<label>', $htmlElement);
            }

            if ($formElement['maxLength'] > 1) {

                if ($jj == $formElement['maxLength']) {
                    $jj = 0;
                    $br = '<br>';
                } else {
                    $br = '';
                }
            }

            $htmlElement = Support::wrapTag("<div class='$radioClass'>", $htmlElement) . $br;

            $html .= $htmlElement;
        }

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * Builds a Selct (Dropdown) Box.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return mixed
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildSelect(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        $attribute = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getAttributeList($formElement, ['autofocus']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        if (isset($formElement['size']) && $formElement['size'] > 1) {
            $attribute .= Support::doAttribute('size', $formElement['size']);
            $attribute .= Support::doAttribute('multiple', 'multiple');
        }

        $option = '';
        $firstSelect = true;
        $jsonValues = array();
        for ($ii = 0; $ii < count($itemValue); $ii++) {
            $option .= '<option ';

            $option .= Support::doAttribute('value', $itemKey[$ii]);

            $jsonValues[] = [
                'value' => $itemKey[$ii],
                'text' => $itemValue[$ii],
                'selected' => ($itemKey[$ii] == $value && $firstSelect)
            ];

            if ($itemKey[$ii] === $value && $firstSelect) {
                $option .= 'selected';
                $firstSelect = false;
            }

            $option .= '>' . $itemValue[$ii] . '</option>';
        }

        $json = $this->getJsonElementUpdate($htmlFormElementId, $jsonValues, $formElement[FE_MODE]);

        return '<select ' . $attribute . '>' . $option . '</select>' . $this->getHelpBlock();
    }

    /**
     * Construct a HTML table of the subrecord data.
     * Column syntax definition: https://wikiit.math.uzh.ch/it/projekt/qfq/qfq-jqwidgets/Documentation#Type:_subrecord
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildSubrecord(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $rcText = false;
        $nameColumnId = 'id';
        $targetTableName = '';
        $flagNew = false;
        $flagEdit = false;
        $flagDelete = false;
        $linkNew = '';
        $control = array();

        $primaryRecord = $this->store->getStore(STORE_RECORD);

        if (!$this->prepareSubrecod($formElement, $primaryRecord, $rcText, $nameColumnId)) {
            return $rcText;
        }

        if (isset($formElement[SUBRECORD_PARAMETER_FORM])) {

            Support::setIfNotSet($formElement, F_EXTRA_DELETE_FORM, '');
            $formElement[F_FINAL_DELETE_FORM] = $formElement[F_EXTRA_DELETE_FORM] != '' ? $formElement[F_EXTRA_DELETE_FORM] : $formElement[SUBRECORD_PARAMETER_FORM];

            $linkNew = Support::wrapTag('<th>', $this->createFormLink($formElement, 0, $primaryRecord, $this->symbol[SYMBOL_NEW], 'New'));

            // Decode settings in subrecordOption
            $flagNew = Support::findInSet(SUBRECORD_NEW, $formElement['subrecordOption']);
            $flagEdit = Support::findInSet(SUBRECORD_EDIT, $formElement['subrecordOption']);
            if ($flagDelete = Support::findInSet(SUBRECORD_DELETE, $formElement['subrecordOption'])) {
                $targetTableName = $this->getFormTable($formElement[SUBRECORD_PARAMETER_FORM]);
            }
        }

        $columns = $linkNew;

        if (isset($formElement['sql1'][0])) {
            // construct column attributes
            $control = $this->getSubrecordColumnControl(array_keys($formElement['sql1'][0]));

            // Subrecord: Column titles
            $columns .= '<th>' . implode('</th><th>', $control['title']) . '</th>';
        }

        if ($flagDelete) {
            $columns .= '<th></th>';
        }

        // Table head
        $html = Support::wrapTag('<tr>', $columns);

        foreach ($formElement['sql1'] as $row) {
            $rowHtml = '';

            if ($flagEdit) {
                $rowHtml .= Support::wrapTag('<td>', $this->createFormLink($formElement, $row[$nameColumnId], $primaryRecord, $this->symbol[SYMBOL_EDIT], 'Edit'));
            } elseif ($flagNew) {
                $rowHtml .= Support::wrapTag('<td>', $rowHtml, false);
            }

            // All columns
            foreach ($row as $columnName => $value) {
                if (isset($control['title'][$columnName])) {
                    $rowHtml .= Support::wrapTag('<td>', $this->renderCell($control, $columnName, $value));
                }
            }

            if ($flagDelete) {
                $toolTip = 'Delete';

                if ($this->showDebugInfo) {
                    $toolTip .= PHP_EOL . "form = '" . $formElement[F_FINAL_DELETE_FORM] . "'" . PHP_EOL . "r = '" . $row[$nameColumnId] . "'";
                }

//                $buttonDelete = $this->buildButtonCode('delete-button', $toolTip, GLYPH_ICON_DELETE, $disabled);

                $s = $this->createDeleteUrl($formElement[F_FINAL_DELETE_FORM], '', $row[$nameColumnId], RETURN_SIP);
//                $rowHtml .= Support::wrapTag('<td>', Support::wrapTag("<button type='button' class='record-delete btn btn-default' data-sip='$s'>", '<span class="glyphicon ' . GLYPH_ICON_DELETE . '"></span>'));
                $rowHtml .= Support::wrapTag('<td>', Support::wrapTag("<button type='button' class='record-delete btn btn-default' data-sip='$s' " . Support::doAttribute('title', $toolTip) . ">", '<span class="glyphicon ' . GLYPH_ICON_DELETE . '"></span>'));


            }

            Support::setIfNotSet($row, FE_SUBRECORD_ROW_CLASS);
            $rowClass = 'record ';
            $rowClass .= $row[FE_SUBRECORD_ROW_CLASS];

            Support::setIfNotSet($row, FE_SUBRECORD_ROW_TITLE);
            $rowTitle = $row[FE_SUBRECORD_ROW_TITLE];

            $rowAttribute = Support::doAttribute('class', $rowClass);
            $rowAttribute .= Support::doAttribute('title', $rowTitle);
            $html .= Support::wrapTag("<tr $rowAttribute>", $rowHtml, true);
        }

        return Support::wrapTag('<table class="table table-hover">', $html, true);
    }

    /**
     * Prepare Subrecord:
     * - check if the current record has an recordId>0. If not, subrecord can't be edited. Return a message.
     * - check if there is an SELECT statement for the subrecords.
     * - determine &$nameColumnId
     *
     * @param $formElement
     * @param $primaryRecord
     * @param $rcText
     * @param $nameColumnId
     * @return bool
     * @throws \qfq\UserFormException
     */
    private function prepareSubrecod(array $formElement, array $primaryRecord, &$rcText, &$nameColumnId) {

        if (!isset($primaryRecord['id'])) {
            $rcText = 'Please save and close record and reopen it.';
            return false;
        }

        if (!is_array($formElement['sql1'])) {
            throw new UserFormException('Missing \'sql1\' Query', ERROR_MISSING_SQL1);
        }

        // No records?
        if (count($formElement['sql1']) == 0) {
            $rcText = '';
            return true;
        }

        // Check if $nameColumnId column exist.
        if (!isset($formElement['sql1'][0][$nameColumnId])) {
            // no: try fallback.
            $nameColumnId = '_id';

            if (!isset($formElement['sql1'][0][$nameColumnId])) {
                throw new UserFormException('Missing column \'id\' or \'_id\' in subrecord query', ERROR_SUBRECORD_MISSING_COLUMN_ID);
            }
        }

        if (!isset($formElement['sql1'][0][$nameColumnId])) {
            throw new UserFormException('Missing column \'id\' (or "_id") in  \'sql1\' Query', ERROR_DB_MISSING_COLUMN_ID);
        }

        return true;
    }

    /**
     * Renders an Link with a symbol (edit/new) and register a new SIP to grant permission to the link..
     *
     * Returns <a href="<Link>">[icon]</a>
     *
     * Link: <page>?s=<SIP>&<standard typo3 params>
     * SIP: form = $formElement['form'] (provided via formElement['parameter'])
     *      r = $targetRecordId
     *      Parse  $formElement['detail'] with possible key/value pairs. E.g.: detail=id:gr_id,#{{a}}:p_id,#12:x_id
     *        gr_id = <<primarytable.id>>
     *        p_id = <<variable defined in SIP or Client>>
     *        x_id = 12 (constant)
     *
     *
     * @param $formElement
     * @param $targetRecordId
     * @param $record
     * @return string
     * @throws UserFormException
     */
    private function createFormLink(array $formElement, $targetRecordId, array $record, $symbol, $toolTip) {

        //TODO: Umstellen auf Benutzung der Link Klasse.

        $queryStringArray = [
            SIP_FORM => $formElement[SUBRECORD_PARAMETER_FORM],
            SIP_RECORD_ID => $targetRecordId
        ];

        // Add custom query parameter
        if (isset($formElement[SUBRECORD_PARAMETER_DETAIL])) {
            $detailParam = KeyValueStringParser::parse($formElement[SUBRECORD_PARAMETER_DETAIL]);
            foreach ($detailParam as $src => $dest) {
                // Constants
                if ($src[0] == '&') {
                    $queryStringArray[$dest] = substr($src, 1);
                    continue;
                }
                // Form record values or parameter
                if (isset($record[$src])) {
                    $queryStringArray[$dest] = $record[$src];
                }
            }
        }

        if ($this->showDebugInfo)
            $toolTip .= PHP_EOL . OnArray::toString($queryStringArray, ' = ', PHP_EOL, "'");

        Support::appendTypo3ParameterToArray($queryStringArray);
        // If there is a specific targetpage defined, take it.
        if (isset($formElement[SUBRECORD_PARAMETER_PAGE]) && $formElement[SUBRECORD_PARAMETER_PAGE] !== '') {
            $queryStringArray['id'] = $formElement[SUBRECORD_PARAMETER_PAGE];
        }

        //-----------------
        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();
        $url = $sip->queryStringToSip($queryString);

        return Support::wrapTag('<a class="btn btn-default" ' . Support::doAttribute('href', $url) . ' title="' . $toolTip . '">', $symbol);
    }

    /**
     * Get the name for the given form $formName. If not found, return ''.
     *
     * @param $formName
     * @return string   tableName for $formName
     * @throws CodeException
     * @throws DbException
     */
    private function getFormTable($formName) {
        $row = $this->db->sql("SELECT " . F_TABLE_NAME . " FROM Form AS f WHERE f.name = ?", ROW_EXPECT_0_1, [$formName]);
        if (isset($row[F_TABLE_NAME])) {
            return $row[F_TABLE_NAME];
        }

        return '';
    }

    /**
     * Get various column format information based on the 'raw' column title. The attributes are separated by '|' and specified as 'key' or 'key=value'.
     *
     * - Return all parsed values as an assoc array.
     * - For regular columns: If there is no 'width' specified, take the default 'SUBRECORD_COLUMN_WIDTH'
     * - For 'icon /  url / mailto': no width limit.
     *
     * Returned assoc array:
     *  title      Only key. Element is non numeric, which is not a keyword 'width/nostrip/icon/url/mailto'
     *  width      Key/Value Pair. Not provided for 'icon/url/mailto'.
     *  nostrip    Only key. Do not strip HTML Tags from the content.
     *  icon       Only key. Value will rendered (later) as an image.
     *  url        Only key. Value will rendered (later) as a 'href'
     *  mailto     Only key. Value will rendered (later) as a 'href mailto'
     *
     *
     * @param array $titleRaw
     * @return array
     * @throws UserFormException
     */
    private function getSubrecordColumnControl(array $titleRaw) {
        $control = array();

        foreach ($titleRaw AS $columnName) {

            switch ($columnName) {
                case FE_SUBRECORD_ROW_CLASS:
                case FE_SUBRECORD_ROW_TITLE:
                    continue 2;
                default:
                    break;
            }

            $flagWidthLimit = true;
            $control['width'][$columnName] = SUBRECORD_COLUMN_WIDTH;

            // a) 'City@width=40', b) 'Status@icon', c) 'Mailto@width=80@nostrip'
            $arr = KeyValueStringParser::parse($columnName, '=', '|', KVP_IF_VALUE_EMPTY_COPY_KEY);
            foreach ($arr as $attribute => $value) {
                switch ($attribute) {
                    case 'width':
                    case 'nostrip':
                    case 'title':
                    case 'link':
                        break;
                    case 'icon':
                    case 'url':
                    case 'mailto':
                        $flagWidthLimit = false;
                        break;
                    default:
                        $attribute = is_numeric($value) ? 'width' : 'title';
                        break;
                }
                $control[$attribute][$columnName] = $value;
            }

            if (!isset($control['title'][$columnName]))
                $control['title'][$columnName] = ''; // Fallback:  Might be wrong, but better than nothing.

            // Don't render Columns starting with '_...'.
            if (substr($control['title'][$columnName], 0, 1) === '_') {
                unset($control['title'][$columnName]); // Do not render column later.
                continue;
            }

            // Limit title length
            $control['title'][$columnName] = substr($control['title'][$columnName], 0, $control['width'][$columnName]);

            if (!$flagWidthLimit) {
                $control['width'][$columnName] = false;
            }

        }
        return $control;
    }

    /**
     * Renders $value as specified in array $control
     *
     * nostrip: by default, HTML tags are removed. With this attribute, the value will be delivered as it is.
     * width: if there is a size limit - apply it.
     * icon: The cell will be rendered as an image. $value should contain the name of an image in 'fileadmin/icons/'
     * mailto: The cell will be rendered as an <a> tag with the 'mailto' attribute.
     * url:  The cell will be rendered as an <a> tag. The value will be exploded by '|'. $value[0] = href, value[1] = text.
     *  E.g. $value = 'www.math.uzh.ch/?id=45&v=234|Show details for Modul 123' >> <a href="www.math.uzh.ch/?id=45&v=234">Show details for Modul 123</a>
     *
     * @param array $control
     * @param $columnName
     * @param $columnValue
     * @return string
     */
    private function renderCell(array $control, $columnName, $columnValue) {
        $link = null;

        switch ($columnName) {
            case FE_SUBRECORD_ROW_CLASS:
            case FE_SUBRECORD_ROW_TITLE:
                return '';
            default:
                break;
        }

        $arr = explode('|', $columnValue);
        if (count($arr) == 1)
            $arr[1] = $arr[0];

        $cell = isset($control['nostrip'][$columnName]) ? $columnValue : strip_tags($columnValue);

        if ($control['width'][$columnName] !== false)
            $cell = substr($cell, 0, $control['width'][$columnName]);

        if (isset($control['icon'][$columnName])) {
            $cell = ($cell === '') ? '' : "<image src='" . PATH_ICONS . "/$cell'>";
        }

        if (isset($control['mailto'][$columnName])) {
            $cell = "<a " . Support::doAttribute('href', "mailto:$arr[0]") . ">$arr[1]</a>";
        }

        if (isset($control['url'][$columnName])) {
            $cell = "<a " . Support::doAttribute('href', $arr[0]) . ">$arr[1]</a>";
        }

        if (isset($control['link'][$columnName])) {
            if (!isset($link)) {
                $link = new Link($this->sip);
            }
            $cell = $link->renderLink($columnValue);
        }

        return $cell;
    }

    /**
     * Create a link (incl. SIP) to delete the current record.
     *
     * @param string $formName if there is a form, specify that
     * @param string $tableName if there is no form , specify the table from where to delete the record.
     * @param int $recordId record to delete
     * @param string $mode
     *  * mode=RETURN_URL: return complete URL
     *  * mode=RETURN_SIP: returns only the sip
     *  * mode=RETURN_ARRAY: returns array with url ('_url') and all decoded and created parameters.
     * @return string String: "API_DIR/delete.php?sip=...."
     * @throws CodeException
     */
    public function createDeleteUrl($formName, $tableName, $recordId, $mode = RETURN_URL) {

        $queryStringArray = [
            SIP_RECORD_ID => $recordId,
            SIP_MODE_ANSWER => MODE_JSON
        ];

        if ($formName !== '') {
            $queryStringArray[SIP_FORM] = $formName;
        }

        if ($tableName !== '') {
            $queryStringArray[SIP_TABLE] = $tableName;
        }

        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();

        return $sip->queryStringToSip($queryString, $mode, API_DIR . '/' . API_DELETE_PHP);
    }

    /**
     * Build an Upload (File) Button.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildFile(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $attribute = '';

        # Build param array for uniq SIP
        $arr = array();
        $arr['fake_uniq_never_use_this'] = uniqid(); // make sure we get a new SIP. This is needed for multiple forms (same user) with r=0
        $arr[CLIENT_SIP_FOR_FORM] = $this->store->getVar(SIP_SIP, STORE_SIP);
        $arr[CLIENT_FE_NAME] = $formElement['name'];
        $arr[CLIENT_FORM] = $this->formSpec['name'];
        $arr[CLIENT_RECORD_ID] = $this->store->getVar(SIP_RECORD_ID, STORE_SIP);
        $arr[CLIENT_PAGE_ID] = 'fake';
        $arr[EXISTING_PATH_FILE_NAME] = $value;
        $sipUpload = $this->sip->queryStringToSip(OnArray::toString($arr), RETURN_SIP);

        $hiddenSipUpload = $this->buildNativeHidden($htmlFormElementId, $sipUpload);

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
//        $attribute .= Support::doAttribute('class', 'form-control');
        $attribute .= Support::doAttribute('type', 'file');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getAttributeList($formElement, ['autofocus', 'accept']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('data-sip', $sipUpload);

        if ($value === '' || $value === false) {
            $textDeleteClass = 'hidden';
            $uploadClass = '';
        } else {
            $textDeleteClass = '';
            $uploadClass = 'hidden';
            $formElement[FE_MODE] = FE_MODE_HIDDEN;
        }

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);
        $attribute .= Support::doAttribute('class', $uploadClass, true);
        $htmlInputFile = '<input ' . $attribute . '>' . $this->getHelpBlock();

        $deleteButton = Support::wrapTag("<button type='button' class='delete-file' data-sip='$sipUpload' name='delete-$htmlFormElementId'>", $this->symbol[SYMBOL_DELETE]);

        $htmlFilename = Support::wrapTag("<span class='uploaded-file-name'>", $value, false);
        $htmlTextDelete = Support::wrapTag("<div class='uploaded-file $textDeleteClass'>", $htmlFilename . ' ' . $deleteButton);

//        <button type="button" class="file-delete" data-sip="571d1fc9e6974"><span class="glyphicon glyphicon-trash"></span></button>

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $htmlTextDelete . $htmlInputFile . $hiddenSipUpload;
    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="date" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws UserFormException
     */
    public function buildDateTime(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {

        $attribute = Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');

        $arrMinMax = null;

        $this->adjustMaxLength($formElement);
        $showTime = ($formElement[FE_TYPE] == 'time' || $formElement[FE_TYPE] == 'datetime') ? 1 : 0;
        $value = Support::convertDateTime($value, $formElement[FE_DATE_FORMAT], $formElement[FE_SHOW_ZERO], $showTime, $formElement[FE_SHOW_SECONDS]);

        $tmpPattern = $formElement['checkPattern'];
        $formElement['checkPattern'] = Support::dateTimeRegexp($formElement[FE_TYPE], $formElement[FE_DATE_FORMAT]);


        switch ($formElement['checkType']) {
            case  SANITIZE_ALLOW_PATTERN:
                $formElement['checkPattern'] = $tmpPattern;
                break;
            case SANITIZE_ALLOW_MIN_MAX_DATE:
                $arrMinMax = explode('|', $formElement['checkPattern'], 2);
                if (count($arrMinMax) != 2) {
                    throw new UserFormException('Missing min|max definition', ERROR_MISSING_MIN_MAX);
                }
                break;
            case SANITIZE_ALLOW_ALL:
            case SANITIZE_ALLOW_ALNUMX:
            case SANITIZE_ALLOW_ALLBUT:
                $formElement['checkType'] = SANITIZE_ALLOW_PATTERN;
                break;
            default:
                throw new UserFormException("Checktype not applicable for date/time: '" . $formElement['checkType'] . "'", ERROR_NOT_APPLICABLE);
        }

        // truncate if necessary
        if ($value != '' && $formElement['maxLength'] > 0) {
            $value = substr($value, 0, $formElement['maxLength']);
        }

        $type = $formElement[FE_TYPE];
        if ($type === 'datetime')
            $type = 'datetime-local';

        $attribute .= Support::doAttribute('type', $type);


        // 'maxLength' needs an upper 'L': naming convention for DB tables!
        $attribute .= $this->getAttributeList($formElement, ['size', 'maxLength']);
        $attribute .= Support::doAttribute('value', htmlentities($value), false);
        $attribute .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        if ($formElement['placeholder'] == '') {
            $timePattern = ($formElement[FE_SHOW_SECONDS] == 1) ? 'hh:mm:ss' : 'hh:mm';
            switch ($formElement[FE_TYPE]) {
                case 'date':
                    $placeholder = $formElement[FE_DATE_FORMAT];
                    break;
                case 'datetime':
                    $placeholder = $formElement[FE_DATE_FORMAT] . ' ' . $timePattern;
                    break;
                case 'time':
                    $placeholder = $timePattern;
                    break;
                default:
                    throw new UserFormException("Unexpected Formelement type: '" . $formElement[FE_TYPE] . "'", ERROR_FORMELEMENT_TYPE);
            }
            $formElement['placeholder'] = $placeholder;
        }

        $attribute .= $this->getAttributeList($formElement, ['autocomplete', 'autofocus', 'placeholder']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getInputCheckPattern($formElement['checkType'], $formElement['checkPattern']);

        if (is_array($arrMinMax)) {
            $attribute .= Support::doAttribute('min', $arrMinMax[0]);
            $attribute .= Support::doAttribute('max', $arrMinMax[1]);
        }

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return "<input $attribute>" . $this->getHelpBlock();

    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="date" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return string
     * @throws UserFormException
     */
    public function buildDateJQW(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $arrMinMax = null;

//        if ($formElement['placeholder'] == '') {
//            $timePattern = ($formElement[FE_SHOW_SECONDS] == 1) ? 'hh:mm:ss' : 'hh:mm';
//            switch ($formElement[FE_TYPE]) {
//                case 'date':
//                    $placeholder = $formElement[FE_DATE_FORMAT];
//                    break;
//                case 'datetime':
//                    $placeholder = $formElement[FE_DATE_FORMAT] . ' ' . $timePattern;
//                    break;
//                case 'time':
//                    $placeholder = $timePattern;
//                    break;
//                default:
//                    throw new UserFormException("Unexpected Formelement type: '" . $formElement[FE_TYPE] . "'", ERROR_FORMELEMENT_TYPE);
//            }
//            $formElement['placeholder'] = $placeholder;
//        }

//        switch ($formElement['checkType']) {
////            case  SANITIZE_ALLOW_PATTERN:
////                $formElement['checkPattern'] = $tmpPattern;
////                break;
//            case SANITIZE_ALLOW_MIN_MAX_DATE:
//                $arrMinMax = explode('|', $formElement['checkPattern'], 2);
//                if (count($arrMinMax) != 2) {
//                    throw new UserFormException('Missing min|max definition', ERROR_MISSING_MIN_MAX);
//                }
//                break;
////            case SANITIZE_ALLOW_ALL:
////            case SANITIZE_ALLOW_ALNUMX:
////            case SANITIZE_ALLOW_ALLBUT:
////                $formElement['checkType'] = SANITIZE_ALLOW_PATTERN;
////                break;
//            default:
//                throw new UserFormException("Checktype not applicable for date/time: '" . $formElement['checkType'] . "'", ERROR_NOT_APPLICABLE);
//        }

        $this->adjustMaxLength($formElement);
        $showTime = ($formElement[FE_TYPE] == 'time' || $formElement[FE_TYPE] == 'datetime') ? 1 : 0;
        $value = Support::convertDateTime($value, $formElement[FE_DATE_FORMAT], $formElement[FE_SHOW_ZERO], $showTime, $formElement[FE_SHOW_SECONDS]);

//        $formElement[FE_DATE_FORMAT]

        $attribute = Support::doAttribute('id', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'jqw-datetimepicker');
        $attribute .= Support::doAttribute('data-control-name', "$htmlFormElementId");
        $attribute .= Support::doAttribute('data-format-string', "dd.MM.yyyy HH:mm");
        $attribute .= Support::doAttribute('data-show-time-button', "true");
//        $attribute .= Support::doAttribute('data-placeholder', $formElement['placeholder']);
        $attribute .= Support::doAttribute('data-value', htmlentities($value), false);
//        $attribute .= Support::doAttribute('data-autofocus', $formElement['autofocus']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('data-title', $formElement['tooltip']);

//        if (is_array($arrMinMax)) {
//            $attribute .= Support::doAttribute('data-min', $arrMinMax[0]);
//            $attribute .= Support::doAttribute('data-max', $arrMinMax[1]);
//        }

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

//        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        $element = Support::wrapTag("<div $attribute>", '', false);

        return $element . $this->getHelpBlock();
    }

    /**
     * Build a HTML 'textarea' element which becomes a TinyMCE Editor.
     * List of possible plugins: https://www.tinymce.com/docs/plugins/
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode
     * @return string
     * @throws \qfq\UserFormException
     */
    public function buildEditor(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {

        //TODO plugin autoresize nutzen um Editorgroesse anzugeben

        $this->adjustMaxLength($formElement);

        $attribute = Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('id', $htmlFormElementId);

        $attribute .= Support::doAttribute('class', 'qfq-tinymce');
        $attribute .= Support::doAttribute('data-control-name', "$htmlFormElementId");
        $attribute .= Support::doAttribute('data-placeholder', $formElement['placeholder']);
//        $attribute .= Support::doAttribute('data-autofocus', $formElement['autofocus']);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('data-title', $formElement['tooltip']);

        $formElement = $this->setEditorConfig($formElement, $htmlFormElementId);
        // $formElement['editor-plugins']='autoresize code'
        // $formElement['editor-contextmenu']='link image | cell row column'
        $json = $this->getPrefixedElementsAsJSON(FE_EDITOR_PREFIX, $formElement);
        $attribute .= Support::doAttribute('data-config', $json, true, ESCAPE_WITH_HTML_QUOTE);


        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);
        $attribute .= $this->getAttributeList($formElement, [F_FE_DATA_PATTERN_ERROR, F_FE_DATA_REQUIRED_ERROR, F_FE_DATA_MATCH_ERROR, F_FE_DATA_ERROR]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        $element = Support::wrapTag("<textarea $attribute>", htmlentities($value), false);

        return $element . $this->getHelpBlock();
    }

    /**
     * Parse $formElement[FE_EDITOR_*] settings and build editor settings.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @return array
     */
    private function setEditorConfig(array $formElement, $htmlFormElementId) {
        $flagMaxHeight = false;

        // plugins
        if (!isset($formElement[FE_EDITOR_PREFIX . 'plugins'])) {
            $formElement[FE_EDITOR_PREFIX . 'plugins'] = 'code link searchreplace table textcolor textpattern visualchars';
        }

        // toolbar: https://www.tinymce.com/docs/advanced/editor-control-identifiers/#toolbarcontrols
        if (!isset($formElement[FE_EDITOR_PREFIX . 'toolbar'])) {
            $formElement[FE_EDITOR_PREFIX . 'toolbar'] = 'code searchreplace undo redo | ' .
                'styleselect link table | fontselect fontsizeselect | ' .
                'bullist numlist outdent indent | forecolor backcolor bold italic';
        }

        // menubar
        if (!isset($formElement[FE_EDITOR_PREFIX . 'menubar'])) {
            $formElement[FE_EDITOR_PREFIX . 'menubar'] = 'false';
        }

        // autofocus
        if (isset($formElement['autofocus']) && $formElement['autofocus'] == 'yes') {
            $formElement[FE_EDITOR_PREFIX . 'auto_focus'] = $htmlFormElementId;
        }

        // Check for min_height, max_height
        $minMax = explode(',', $formElement['size'], 2);
        if (isset($minMax[0]) && ctype_digit($minMax[0]) && !isset($formElement[FE_EDITOR_PREFIX . 'min_height'])) {
            $formElement[FE_EDITOR_PREFIX . 'min_height'] = $minMax[0];
        }
        if (isset($minMax[1]) && ctype_digit($minMax[1]) && !isset($formElement[FE_EDITOR_PREFIX . 'max_height'])) {
            $formElement[FE_EDITOR_PREFIX . 'max_height'] = $minMax[1];
            $flagMaxHeight = true;
        }

        // statusbar: disable if not user defined and if no max_height is given.
        if (!$flagMaxHeight && !isset($formElement[FE_EDITOR_PREFIX . 'statusbar'])) {
            $formElement[FE_EDITOR_PREFIX . 'statusbar'] = 'false';
        }

        return $formElement;
    }

    /**
     * Searches for '$prefix*' elements in $formElement. Collect all found elements, strip $prefix (=$keyName) and
     *   returns keys/values JSON encoded.
     * Only 'alpha' chars are allowed as keyName.
     * Empty $settings are ok.
     *
     * @param string $prefix
     * @param array $formElement
     * @return string
     * @throws \qfq\UserFormException
     */
    private function getPrefixedElementsAsJSON($prefix, array $formElement) {
        $settings = array();

        // E.g.: $key = editor-plugins
        foreach ($formElement as $key => $value) {
            if (substr($key, 0, strlen($prefix)) == $prefix) {

                $keyName = substr($key, strlen($prefix));

                if ($keyName == '') {
                    throw new UserFormException("Empty '" . $prefix . "*' keyname: '" . $keyName . "'", ERROR_INVALID_EDITOR_PROPERTY_NAME);
                }

                // $value might be boolean false, which should be used! Do not compare with ''.
                if (isset($value)) {
                    // real boolean are important for TinyMCE config 'statusbar', 'menubar', ...
                    if ($value === 'false') $value = false;
                    if ($value === 'true') $value = true;
                    $settings[$keyName] = $value;
                }
            }
        }

        return json_encode($settings);
    }

    /**
     * Build Grid JQW element.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param $fake
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @throws UserFormException
     */
    public function buildGridJQW(array $formElement, $htmlFormElementId, $value, $fake, $mode = FORM_LOAD) {
        // TODO: implement
        throw new UserFormException("Not implemented yet: buildGridJQW()", ERROR_NOT_IMPLEMENTED);
    }

    /**
     * Build Note.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @param array $json
     * @return mixed
     */
    public function buildNote(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        return $value;
    }

    /**
     * Build Pill:
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return mixed
     */
    public function buildPill(array $formElement, $htmlFormElementId, $value, array &$json) {
        return $value;
    }

    /**
     * Build a HTML fieldset. Renders all assigned FormElements inside the fieldset.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return mixed
     * @throws CodeException
     * @throws DbException
     */
    public function buildFieldset(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $attribute = '';

        // save parent processed FE's
        $tmpStore = $this->feSpecNative;

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('data-load', ($formElement[FE_DYNAMIC_UPDATE] === 'yes') ? 'data-load' : '');

        // <fieldset>
        $html = '<fieldset ' . $attribute . '>';


        if ($formElement[FE_LABEL] !== '') {
            $html .= '<legend>' . $formElement[FE_LABEL] . '</legend>';
        }

        $html .= $this->wrap[WRAP_SETUP_IN_FIELDSET][WRAP_SETUP_START];

        // child FE's
        $this->feSpecNative = $this->db->getNativeFormElements(SQL_FORM_ELEMENT_SPECIFIC_CONTAINER,
            ['yes', $this->formSpec["id"], 'native,container', $formElement['id']], $this->formSpec);

        $html .= $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), FORM_ELEMENTS_NATIVE_SUBRECORD, 0, $json);

        $html .= $this->wrap[WRAP_SETUP_IN_FIELDSET][WRAP_SETUP_END];

        $html .= '</fieldset>';

        // restore parent processed FE's
        $this->feSpecNative = $tmpStore;

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * Build a 'templateGroup'. Renders all assigned FormElements of the templateGroup.
     * If there are already vlaues for the formElements, fill as much copies as values exist
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     * @return mixed
     * @throws CodeException
     * @throws DbException
     */
    public function buildTemplateGroup(array $formElement, $htmlFormElementId, $value, array &$json, $mode = FORM_LOAD) {
        $attribute = '';
        $html = '';

        $addClass = Support::setIfNotSet($formElement, FE_TEMPLATE_GROUP_ADD_CLASS, 'btn btn-default');
        $addText = Support::setIfNotSet($formElement, FE_TEMPLATE_GROUP_ADD_TEXT, 'Add');
        $removeClass = Support::setIfNotSet($formElement, FE_TEMPLATE_GROUP_REMOVE_CLASS, 'btn btn-default');
        $removeText = Support::setIfNotSet($formElement, FE_TEMPLATE_GROUP_REMOVE_TEXT, 'Remove');
        $classCustom = Support::setIfNotSet($formElement, FE_TEMPLATE_GROUP_CLASS);

        // save parent processed FE's
        $feSpecNativeSave = $this->feSpecNative;

        $qfqFieldsName = 'qfq_fields_' . $formElement[FE_ID]; // ='qfq-fields'
        $templateName = 'template_' . $formElement[FE_ID]; // ='template'
        $targetName = 'target_' . $formElement[FE_ID]; // ='template'
        Support::setIfNotSet($formElement, FE_MAX_LENGTH, '5', '');
        $max = $formElement[FE_MAX_LENGTH];

        $codeJs = <<<EOT
<script type="text/javascript">
    $(function () {
        $(".$qfqFieldsName").each(
            function () {
                QfqNS.initializeFields(this);
            }
        );
    });
</script>
EOT;

        $htmlAdd = <<<EOT
<button type="button" class="$addClass" onclick="QfqNS.addFields('#$templateName', '#$targetName', $max)">$addText</button>
EOT;

        $htmlDelete = <<<EOT
<div class="qfq-delete-button">
    <button type="button"  class="$removeClass" onclick="QfqNS.removeFields(this)">$removeText</button>
</div>
EOT;

        // child FE's
        $this->feSpecNative = $this->db->getNativeFormElements(SQL_FORM_ELEMENT_SPECIFIC_CONTAINER,
            ['yes', $this->formSpec["id"], 'native,container', $formElement['id']], $this->formSpec);

        // Count defined FormElements in the current templateGroup
        $elementsTotal = count($this->feSpecNative);
        if ($elementsTotal < 1) {
            $this->feSpecNative = $feSpecNativeSave;
            return '';
        }

        // If there are already elements filled, take them.
        $html = $this->templateGroupCollectFilledElements($max, $htmlDelete, $json);

        $attribute = Support::doAttribute('class', $qfqFieldsName);
        $attribute .= Support::doAttribute('id', $targetName);
        $attribute .= Support::doAttribute('data-qfq-line-template', '#' . $templateName);

        // Element where the effective FormElements will be copied to. The BS 'col-md-* Classes are inside the template, not here. This here should be pure data without wrapping.
        $html = Support::wrapTag("<div $attribute>", $html, false);

        // Add button, row below: The label & note of the FormElement 'templateGroup' will be used for the add button row.
        $tmpFe = $formElement;
        $tmpFe[FE_NAME] = '_add';
        $tmpFe[FE_LABEL] = '';
        $tmpFe[FE_NOTE] = '';
        $this->fillWrapLabelInputNote($tmpFe[FE_BS_LABEL_COLUMNS], $tmpFe[FE_BS_INPUT_COLUMNS], $tmpFe[FE_BS_NOTE_COLUMNS]);
        $html .= $this->buildRowNative($tmpFe, $htmlAdd, $htmlFormElementId);

        $html = $this->wrap[WRAP_SETUP_IN_TEMPLATE_GROUP][WRAP_SETUP_START] . $html . $this->wrap[WRAP_SETUP_IN_TEMPLATE_GROUP][WRAP_SETUP_END];

        // Append 'delete' Button after 'note' of the templateGroups last element.
        $this->feSpecNative[$elementsTotal - 1][FE_NOTE] .= $htmlDelete;

        // Get FE natives
        $template = $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), FORM_ELEMENTS_NATIVE, 0, $json);

        // Wrap the main html code
        $attribute = Support::doAttribute('class', ['qfq-line', $classCustom]);
        $template = Support::wrapTag("<div $attribute>", $template);
        $template = Support::wrapTag('<script id="' . $templateName . '" type="text/' . $templateName . '">', $template);

        $html .= $template . $codeJs;

        // restore parent processed FE's
        $this->feSpecNative = $feSpecNativeSave;

        //TODO: nicht klar ob das hier noetig ist.
//        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $html;

    }

    /**
     * Build real FormElements based on the current templateGroup FormElements in $this->feSpecNative.
     * Take a templateGroup 'copy', if at least one of the elements is filled with a non default value.
     * Stop filling elements if there are no further elements filled.
     *
     * @param int $max
     * @param string $htmlDelete
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    private function templateGroupCollectFilledElements($max, $htmlDelete, array &$json) {

        $record = $this->store->getStore(STORE_RECORD); // current values
        if ($record === false || count($record) === 0) {
            return '';
        }
        $default = $this->store->getStore(STORE_TABLE_DEFAULT); // current defaults

        $lastFilled = 0; // Marker if there is at least one element per copy who is filled.
        $feSpecNativeCopy = array();
        for ($ii = 1; $ii < $max; $ii++) {

            // Per copy, iterate over all templateGroup FormElements
            foreach ($this->feSpecNative as $fe) {

                $columnName = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_NAME]);

                $fe[FE_LABEL] = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_LABEL]);
                $fe[FE_NOTE] = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_NOTE]);
                if (isset($record[$columnName])) {
                    if ($record[$columnName] != $default[$columnName]) {
                        $lastFilled = $ii;
                    }
                } else {
                    throw new UserFormException("Not implemented: templateGroup FormElement-columns not found in primary table.", ERROR_NOT_IMPLEMENTED);
                }
                $fe[FE_NAME] = $columnName;
                $feSpecNativeCopy[$ii - 1][] = $fe; // Build array with current copy of templateGroup.
            }

            // Append $htmlDelete on the last element of all copies,, but not the first.
            if ($ii > 1) {
                // Count defined FormElements in the current templateGroup
                $last = count($feSpecNativeCopy[$ii - 1]) - 1;
                // Append 'delete' Button at the note of the last element
                $feSpecNativeCopy[$ii - 1][$last][FE_NOTE] .= $htmlDelete;
            }
        }

        // Nothing found: return
        if (count($feSpecNativeCopy) == 0) {
            return '';
        }

        $feSpecNativeSave = $this->feSpecNative;

        $html = '';
        for ($ii = 0; $ii < $lastFilled; $ii++) {
            $this->feSpecNative = $feSpecNativeCopy[$ii];
            $htmlCopy = $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), FORM_ELEMENTS_NATIVE, 0, $json);
            $htmlCopy = Support::wrapTag('<div class="qfq-line">', $htmlCopy);
            $html .= $htmlCopy;
        }

        $this->feSpecNative = $feSpecNativeSave;

        return $html;
    }

    abstract public function buildRowNative(array $formElement, $htmlElement, $htmlFormElementId);


    /**
     * Create a delete link.
     *
     * @param $table
     * @param $recordId
     * @param $symbol
     * @param $toolTip
     * @return string
     */
//    private function createDeleteLink($table, $recordId, $symbol, $toolTip) {
//
//        if ($this->showDebugInfo) {
//            $toolTip .= PHP_EOL . "table = '$table'" . PHP_EOL . "id = '$recordId'";
//        }
//
//        $url = $this->createDeleteUrl('', $table, $recordId);
//
//        return Support::wrapTag('<a ' . Support::doAttribute('href', $url) . ' title="' . $toolTip . '">', $symbol);
//
//    }

}