<?php
/**
 * Created by PhpStorm.
 * @author: crose
 * Date: 7/8/15
 * Time: 10:12 AM
 *
 * $Id$
 */

namespace qfq;

use qfq;
use qfq\UserFormException;

require_once(__DIR__ . '/../exceptions/UserFormException.php');
require_once(__DIR__ . '/../../qfq/Constants.php');

/**
 * Class KeyValueStringParser
 *
 * KeyValueStringParser is a parser for strings of the form
 *
 *     key1<delimiterA>value1<delimiterB>key2<delimiterA>value2<delimiterB>
 *
 * For instance
 *
 *     id=1,name=doe,firstname=john
 *
 * - Leading and trailing whitespace will be removed from keys and values.
 * - If a value is surrounded by quotes (`'`,  `"`), leading and trailing
 *   whitespace will be preserved (leading/trailing, quotes will be removed from value).
 * - Comments, lines starting with a '#' or ';', will be skipped.
 *
 * @package qfq
 */
class KeyValueStringParser {

    /**
     * Builds a string based on kvp array. Concatenatet by the given delimiter.
     *
     * @param array $keyValueArray
     * @param string $keyValueDelimiter
     * @param string $listDelimiter
     * @return string
     */
    public static function unparse(array $keyValueArray, $keyValueDelimiter = ":", $listDelimiter = ",") {
        array_walk($keyValueArray, function (&$value) {
            if (!is_string($value) || $value === "" || strlen($value) === 1) {
                return;
            }

            if ($value[0] === " " && self::isFirstAndLastCharacterIdentical($value[0])) {
                $value = '"' . $value . '"';
            }
        });

        $newKeyValuePairImploded = array();
        foreach ($keyValueArray as $key => $value) {
            $newKeyValuePairImploded[] = trim($key) . $keyValueDelimiter . $value;
        }

        return implode($listDelimiter, $newKeyValuePairImploded);
    }

    /**
     * @param $string
     * @return bool
     */
    private static function isFirstAndLastCharacterIdentical($string) {
        if ($string === "") {
            return false;
        }

        return $string[0] === $string[strlen($string) - 1];
    }

    /**
     * Parse key/value pairs string and returns them as an assoc array
     *
     * Hint $keyValueString: "a:1,b:2,c:,d",  "," (empty key AND empty value)
     *
     * @param string $keyValueString string of key/value pairs. E.g.: 'a=100,b=test'
     * @param string $keyValueDelimiter
     * @param string $listDelimiter
     * @param string $valueMode
     *  * KVP_VALUE_GIVEN: If only a key is given, the value is ''.  E.G. 'a,b' >> [ 'a' => '', 'b' => '' ]
     *  * KVP_IF_VALUE_EMPTY_COPY_KEY: If only a key is given, the value is the same as the key. E.G. 'a,b' >> [ 'a' => 'a', 'b' => 'b' ].
     * @return array  associative array indexed by keys
     * @throws UserFormException Thrown if there is a value but no key.
     */

    public static function parse($keyValueString, $keyValueDelimiter = ":", $listDelimiter = ",", $valueMode = KVP_VALUE_GIVEN) {
        if ($keyValueString === "") {
            return array();
        }

        // Clean any "\r\n" to "\n"
        $keyValueString = str_replace("\r\n", "\n", $keyValueString);

        $keyValuePairs = explode($listDelimiter, $keyValueString);

        $returnValue = array();
        foreach ($keyValuePairs as $keyValuePairString) {

            if (trim($keyValuePairString === "")) {
                continue;
            }

            $keyValueArray = explode($keyValueDelimiter, $keyValuePairString, 2);

            $key = trim($keyValueArray[0]);

            // skip comments
            if (substr($key, 0, 1) == '#' || substr($key, 0, 1) == ';') {
                continue;
            }

            if ($key === '') {
                // ":", ":1"
                throw new UserFormException("Value has no key: '$keyValuePairString'", ERROR_KVP_VALUE_HAS_NO_KEY);
            }

            if (count($keyValueArray) === 2) {
                // "a:1", "a:"
                $returnValue[$key] = self::removeSourroundingQuotes(trim($keyValueArray[1]));
            } else {
                // no Value given: "a"
                $returnValue[$key] = ($valueMode === KVP_VALUE_GIVEN) ? "" : $key;
            }
        }

        return $returnValue;
    }

    /**
     * @param $string
     * @return string
     */
    private static function removeSourroundingQuotes($string) {
        $quotes = ['\'', '"'];

        if ($string === "" || strlen($string) === 1) {
            return $string;
        }

        if (in_array($string[0], $quotes) === true && self::isFirstAndLastCharacterIdentical($string)) {
            return substr($string, 1, strlen($string) - 2);
        }

        return $string;
    }
}
